#!/usr/bin/perl
use strict;
use warnings;
use Fcntl qw(:flock);

my $OK             = 0; # Alles OK
my $BADPARAM       = 1; # ungueltige Parameter uebergeben
my $NOTFOUND       = 2; # plattenid nicht existent
my $UNMOUNTFAILED  = 3; # platte kann gefunden aber nicht entlanden werden
my $ALREADYRUNNING = 4; # programm luft schon
my $INTERNALERROR  = 5; # fataler Programmfehler und abbruch
my $STILLRUNNING   = 6; # Disk already running

my $MAXID      = 20;
my $MAXWAIT    = 30;
my $MAXDISWAIT = 10;
my $setfile    = "/dev/usb/iowarrior0";
my $getfile    = "/dev/usb/iowarrior1";
my $statusfile = "/tmp/iowarriorstatus";
my $lockfile   = "/tmp/.iowarrior.lock";
my $MAXTRY     = 30; # Sekunden
my $exit       = $OK;

unless (open(LOCKFILE, ">", $lockfile)) {
   my $err = "Cannot open locking file '".$lockfile."' :".$!;
   syswrite(STDOUT, $err."\n");
   exit($INTERNALERROR);
}

sub doLocking {
   my $noprint = shift;
   my $loop = 0;
   while (1) {
      if (flock(LOCKFILE, LOCK_EX | LOCK_NB)) {
         print STDERR "\n" if $loop;
         last;
      }
      print STDERR scalar(time())." ".$$." Already locked... retrying...\r";
      sleep(1);
      if ($loop++ > $MAXTRY) {
         print STDERR "\n".scalar(time())." Aborting.\n";
         $exit ||= $ALREADYRUNNING;
      }
   }
}

doLocking();

unless (-e $setfile) {
   syswrite(STDOUT, "No set driver found on iowarrior\n");
   exit($INTERNALERROR);
}
unless (-e $getfile) {
   syswrite(STDOUT, "No get driver found on iowarrior\n");
   exit($INTERNALERROR);
}
unless (open(SET,   "+<", $setfile)) {
   syswrite(STDOUT, "Can't open driver 1 on iowarior: ".$!."\n");
   exit($INTERNALERROR);
}
unless (open(GET,   "+<", $getfile)) {
   syswrite(STDOUT, "Can't open driver 2 on iowarior: ".$!."\n");
   exit($INTERNALERROR);
}

my $curstatus = getArrayStatus();
#syswrite(STDOUT, "Flag Current:".DWORDtobit($curstatus)."\n");

my $idcache = [];

if (-e $statusfile) {
   unless (open(STATUS, "<", $statusfile)) {
      syswrite(STDOUT, "Can't open status file: ".$!."\n");
      exit($INTERNALERROR);
   }
   while(<STATUS>) {
      chomp;
      my $split = [split(/;/, $_, 3)];
      if ($split && (ref($split) eq "ARRAY") && (scalar(@$split) == 3) && ($split->[0] =~ /^\d+$/) && ($split->[0] < $MAXID)) {
         if ($split->[1] =~ /^sd[a-z]$/) {
            if (-e "/dev/".$split->[1]) {
               $idcache->[$split->[0]] = [$split->[1], $split->[2]];
            } else {
               syswrite(STDOUT, "Turning off ".$split->[0]." cause device cannot be found in /dev/!\n");
               $exit ||= $NOTFOUND;
               $idcache->[$split->[0]] = undef;
            }
         } else {
            syswrite(STDOUT, "Disk ".$split->[0]." has bad device id ".$split->[1].": Turning off and deleting entry.\n");
            $exit ||= $BADPARAM;
         }
      }
   }
}

writeIDFile();

my $pre = $curstatus;
for(my $platte = 1; $platte < $MAXID; $platte++) {
   my $flagplatte  = 2 ** ($platte-1);
   my $rflagplatte = $flagplatte ^ (unpack("L", chr(0xFF) x 4));
   unless ($curstatus & $flagplatte) {
      unless (defined($idcache->[$platte]) &&
                 (ref($idcache->[$platte]) eq "ARRAY") &&
                      $idcache->[$platte]->[0]) {
         syswrite(STDOUT, "Turning off ".$platte." cause of invalid or missing definition.\n");
         $exit ||= $BADPARAM;
         $curstatus |= $flagplatte;
      }
   }
}

if ($pre != $curstatus) {
   setStatus($curstatus);
   syswrite(STDOUT, "Waiting ".$MAXDISWAIT." seconds to settle");
   for(my $i = 0; $i < $MAXDISWAIT; $i++) {
      syswrite(STDOUT, ".");
      sleep(1);
   }
   syswrite(STDOUT, "\n");
}

my $ret = undef;
sub getArrayStatus {
   my $request = pack("Q", 0xFF);   
   if ($ret = syswrite(GET, $request, length($request)) != length($request)) {
      syswrite(STDOUT, "ERROR REQUESTING STATUS: SENT ONLY ".$ret." BYTES OF ".length($request)."\n");
      exit($INTERNALERROR);
   }
   my $in;
   if ($ret = sysread(GET, $in, 8) != 8) {
      syswrite(STDOUT, "ERROR REQUESTING STATUS: READ ONLY ".$ret." BYTES OF 8\n");
      exit($INTERNALERROR);
   }
   my $values = [split(//, $in)];
   return unpack("L", $values->[1].$values->[2].$values->[3].$values->[4]);
}

sub setStatus {
   my $request = pack("L", shift);
   if ($ret = syswrite(SET, $request, length($request)) != length($request)) {
      syswrite(STDOUT, "ERROR REQUESTING STATUS: SENT ONLY ".$ret." BYTES OF ".length($request)."\n");
      exit($INTERNALERROR);
   }
}

sub DWORDtobit {
   return join("", map { unpack("B8",$_) } reverse split(//, pack("L", shift)));
}

sub getSDs {
   my $return = [];
   unless (opendir(DIR, "/dev/")) {
      syswrite(STDOUT, "Can't open /dev/ for listing!\n");
      exit($INTERNALERROR);
   }
   while($_ = readdir(DIR)) {
      next unless m,^sd[a-z]$,;
      chomp;
      push(@$return, $_);
   }
   close DIR;
   return $return;
}

sub writeIDFile {
   unless (open(STATUS, ">", $statusfile)) {
      syswrite(STDOUT, "Can't open status file: ".$!."\n");
      exit($INTERNALERROR);
   }
   my $id = 0;
   foreach my $entry (@$idcache) {
      print STATUS $id.";".$entry->[0].";".$entry->[1]."\n" if (defined($entry) && (ref($entry) eq "ARRAY"));
      $id++;
   }
}

sub searchForMount {
   my $name = shift;
   my $return = [];
   unless (open(MOUNTS, "<", "/proc/mounts")) {
      syswrite(STDOUT, "Can't open /proc/mounts for reading!\n");
      exit($INTERNALERROR);
   }
   while(<MOUNTS>) {
      chomp;
      next unless m,$name,;
      push(@$return, $_);
   }
   return $return;
}

if (my $platte = $ARGV[0]) {
   if ($platte !~ /^\d+$/) {
      my $id = 0;
      foreach my $curdisk (@$idcache) {
         if (defined($curdisk) &&
                (ref($curdisk) eq "ARRAY") &&
                    ($curdisk->[1]) &&
                    ($curdisk->[1] eq $ARGV[0])) {
            $platte = $id;
            last;
         }
         $id++;
      }
      if ($platte !~ /^\d+$/) {
         syswrite(STDOUT, "Unknown mount point: ".$ARGV[0]."\n");
         exit($INTERNALERROR);
      }
   }
   if (($platte =~ /^\d+$/) && ($platte > 0) && ($platte < $MAXID)) {
      my $flagplatte  = 2 ** ($platte-1);
      my $rflagplatte = $flagplatte ^ (unpack("L", chr(0xFF) x 4));
      #syswrite(STDOUT, " Flag Platte:".DWORDtobit($flagplatte)."\n");
      #syswrite(STDOUT, "RFlag Platte:".DWORDtobit($rflagplatte)."\n");
      if (!defined($ARGV[1]) || ($ARGV[1] =~ /^off$/) || ($ARGV[1] eq "")) {
         if ($curstatus & $flagplatte) {
            syswrite(STDOUT, "Disk ".$platte." not running.\n");
            $exit ||= $NOTFOUND;
         } else {
            syswrite(STDOUT, "Turning ".$platte." off.");
            my $curmounts = [];
            my $waittime = 0;
            if (defined($idcache->[$platte]) &&
                   (ref($idcache->[$platte]) eq "ARRAY") &&
                        $idcache->[$platte]->[0]) {
               $curmounts = searchForMount($idcache->[$platte]->[0]);
               if (@$curmounts) {
                  syswrite(STDOUT, "\nUnmounting ".scalar(@$curmounts)." Mountpoints");
                  while(scalar(@$curmounts) && ($waittime++ <= $MAXWAIT)) {
                     foreach my $mount (@$curmounts) {
                        my $split = [split(/\s/, $mount)];
                        system("/bin/umount ".$split->[0]);
                        rmdir($split->[0]);
                     }
                     syswrite(STDOUT, ".");
                     sleep(1);
                     $curmounts = searchForMount($idcache->[$platte]->[0]);
                  }
               }
            }
            syswrite(STDOUT, "\n");
            if (scalar(@$curmounts)) {
               syswrite(STDOUT, "BAD: DISK STILL MOUNTED: ".join(",", @$curmounts)."\n");
               $exit ||= $UNMOUNTFAILED;
            } else {
               $curstatus |= $flagplatte;
               setStatus($curstatus);
               $idcache->[$platte] = undef;
               writeIDFile();
               syswrite(STDOUT, "Waiting ".$MAXDISWAIT." seconds to settle");
               for(my $i = 0; $i < $MAXDISWAIT; $i++) {
                  syswrite(STDOUT, ".");
                  sleep(1);
               }
               syswrite(STDOUT, "\n");
            }
         }
      } elsif ($ARGV[1]) {
         if ($curstatus & $flagplatte) {
            syswrite(STDOUT, "Turning ".$platte." on");
            rmdir($ARGV[1])
               if ($ARGV[1] !~ /^on$/);
            my $preSDs = getSDs();
            $curstatus &= $rflagplatte;
            setStatus($curstatus);
            my $waittime = 0;
            my $found = [];
            while($waittime++ <= $MAXWAIT) {
               my $curSDs = getSDs();
               $found = [grep { my $curdir = $_; !scalar(grep { $_ eq $curdir } @$preSDs) } @$curSDs];
               if (@$found) {
                  syswrite(STDOUT, " OK:".join(",", @$found));
                  last;
               }
               $preSDs = $curSDs;
               syswrite(STDOUT, ".");
               sleep(1);
            }
            syswrite(STDOUT, "\n");
            if (scalar(@$found) == 1) {
               $idcache->[$platte] = [$found->[0], $ARGV[1]];
               if ($ARGV[1] !~ /^on$/) {
                  mkdir($ARGV[1]);
                  #syswrite(STDOUT, "/bin/mount /dev/".$idcache->[$platte]->[0]."1 ".$idcache->[$platte]->[1]."\n");
                  sleep(1);
                  system("/bin/mount /dev/".$idcache->[$platte]->[0]."1 ".$idcache->[$platte]->[1]);
               }
            } else {
               syswrite(STDOUT, "BAD: DEV NOT FOUND FOR ".$platte.": Turning disk back to off!\n");
               $exit ||= $NOTFOUND;
               $curstatus |= $flagplatte;
               setStatus($curstatus);
               $idcache->[$platte] = undef;
            }
            writeIDFile();
         } else {
            syswrite(STDOUT, "Disk ".$platte." already running.\n");
            #$exit ||= $STILLRUNNING;
         }
      }
      #syswrite(STDOUT, "Flag Current:".DWORDtobit($curstatus)."\n");
   } else {
      syswrite(STDOUT, "Bad format for platte (must be numeric): ".$platte."\n");
      $exit ||= $BADPARAM;
   }
}

exit($exit);
